@@ -1,36 +1,42 @@
+from __future__ import annotations
+
 import re
 import string
-from functools import partial
+from typing import Dict
+from typing import Optional
 
 from pycountry import countries
+from pycountry.db import Data
 
 from schwifty import common
 from schwifty import exceptions
 from schwifty import registry
 from schwifty.bic import BIC
 
-_spec_to_re = {"n": r"\d", "a": r"[A-Z]", "c": r"[A-Za-z0-9]", "e": r" "}
+_spec_to_re: Dict[str, str] = {"n": r"\d", "a": r"[A-Z]", "c": r"[A-Za-z0-9]", "e": r" "}
 
-_alphabet = string.digits + string.ascii_uppercase
+_alphabet: str = string.digits + string.ascii_uppercase
 
 
-def _get_iban_spec(country_code):
+def _get_iban_spec(country_code: str) -> dict:
     try:
-        return registry.get("iban")[country_code]
+        spec = registry.get("iban")
+        assert isinstance(spec, dict)
+        return spec[country_code]
     except KeyError:
         raise exceptions.InvalidCountryCode(f"Unknown country-code '{country_code}'")
 
 
-def numerify(string):
+def numerify(string: str) -> int:
     return int("".join(str(_alphabet.index(c)) for c in string))
 
 
-def code_length(spec, code_type):
+def code_length(spec: Dict, code_type: str) -> int:
     start, end = spec["positions"][code_type]
     return end - start
 
 
-def _calc_it_checksum(bban):
+def _calc_it_checksum(bban: str) -> str:
     odds = [
         1,
         0,
@@ -71,7 +77,7 @@ def _calc_it_checksum(bban):
     return _alphabet[sum_ % 26 + 10]
 
 
-def add_bban_checksum(country_code, bban):
+def add_bban_checksum(country_code: str, bban: str) -> str:
     if country_code == "IT":
         checksum = _calc_it_checksum(bban[1:])
         bban = checksum + bban[1:]
@@ -109,7 +115,7 @@ class IBAN(common.Base):
         InvalidLength: If the length does not match the country specific specification.
     """
 
-    def __init__(self, iban, allow_invalid=False):
+    def __init__(self, iban: str, allow_invalid: Optional[bool] = False) -> None:
         super().__init__(iban)
         if self.checksum_digits == "??":
             self._code = self.country_code + self._calc_checksum_digits() + self.bban
@@ -117,11 +123,13 @@ class IBAN(common.Base):
         if not allow_invalid:
             self.validate()
 
-    def _calc_checksum_digits(self):
+    def _calc_checksum_digits(self) -> str:
         return "{:02d}".format(98 - (numerify(self.bban + self.country_code) * 100) % 97)
 
     @classmethod
-    def generate(cls, country_code, bank_code, account_code, branch_code=""):
+    def generate(
+        cls, country_code: str, bank_code: str, account_code: str, branch_code: str = ""
+    ) -> IBAN:
         """Generate an IBAN from it's components.
 
         If the bank-code and/or account-number have less digits than required by their
@@ -146,10 +154,10 @@ class IBAN(common.Base):
             Added the `branch_code` parameter to allow the branch code (or sort code) to be
             specified independently.
         """
-        spec = _get_iban_spec(country_code)
-        bank_code_length = code_length(spec, "bank_code")
-        branch_code_length = code_length(spec, "branch_code")
-        account_code_length = code_length(spec, "account_code")
+        spec: Dict = _get_iban_spec(country_code)
+        bank_code_length: int = code_length(spec, "bank_code")
+        branch_code_length: int = code_length(spec, "branch_code")
+        account_code_length: int = code_length(spec, "account_code")
 
         if len(bank_code) == bank_code_length + branch_code_length:
             bank_code, branch_code = bank_code[:bank_code_length], bank_code[bank_code_length:]
@@ -182,7 +190,7 @@ class IBAN(common.Base):
         bban = add_bban_checksum(country_code, bban)
         return cls(country_code + "??" + bban)
 
-    def validate(self):
+    def validate(self) -> bool:
         """Validate the structural integrity of this IBAN.
 
         This function will verify the country specific format as well as the Luhn checksum.
@@ -203,19 +211,19 @@ class IBAN(common.Base):
         self._validate_checksum()
         return True
 
-    def _validate_characters(self):
+    def _validate_characters(self) -> None:
         if not re.match(r"[A-Z]{2}\d{2}[A-Z]*", self.compact):
             raise exceptions.InvalidStructure(f"Invalid characters in IBAN {self.compact}")
 
-    def _validate_checksum(self):
+    def _validate_checksum(self) -> None:
         if self.numeric % 97 != 1 or self._calc_checksum_digits() != self.checksum_digits:
             raise exceptions.InvalidChecksumDigits("Invalid checksum digits")
 
-    def _validate_length(self):
+    def _validate_length(self) -> None:
         if self.spec["iban_length"] != self.length:
             raise exceptions.InvalidLength("Invalid IBAN length")
 
-    def _validate_format(self):
+    def _validate_format(self) -> None:
         if not self.spec["regex"].match(self.bban):
             raise exceptions.InvalidStructure(
                 "Invalid BBAN structure: '{}' doesn't match '{}''".format(
@@ -224,7 +232,7 @@ class IBAN(common.Base):
             )
 
     @property
-    def is_valid(self):
+    def is_valid(self) -> bool:
         """bool: Indicate if this is a valid IBAN.
 
         Note:
@@ -243,22 +251,22 @@ class IBAN(common.Base):
             return False
 
     @property
-    def numeric(self):
+    def numeric(self) -> int:
         """int: A numeric represenation of the IBAN."""
         return numerify(self.bban + self.compact[:4])
 
     @property
-    def formatted(self):
+    def formatted(self) -> str:
         """str: The IBAN formatted in blocks of 4 digits."""
         return " ".join(self.compact[i : i + 4] for i in range(0, len(self.compact), 4))
 
     @property
-    def spec(self):
+    def spec(self) -> Dict:
         """dict: The country specific IBAN specification."""
         return _get_iban_spec(self.country_code)
 
     @property
-    def bic(self):
+    def bic(self) -> Optional[BIC]:
         """BIC or None: The BIC associated to the IBAN´s bank-code.
 
         If the bank code is not available in Schwifty's registry ``None`` is returned.
@@ -269,47 +277,53 @@ class IBAN(common.Base):
         try:
             return BIC.from_bank_code(self.country_code, self.bank_code or self.branch_code)
         except exceptions.SchwiftyException:
-            pass
+            return None
 
     @property
-    def country(self):
+    def country(self) -> Optional[Data]:
         """Country: The country this IBAN is registered in."""
         return countries.get(alpha_2=self.country_code)
 
-    def _get_code(self, code_type):
+    def _get_code(self, code_type: str) -> str:
         start, end = self.spec["positions"][code_type]
         return self.bban[start:end]
 
-    bban = property(
-        partial(common.Base._get_component, start=4),
-        doc="str: The BBAN part of the IBAN.",
-    )
-    country_code = property(
-        partial(common.Base._get_component, start=0, end=2),
-        doc="str: ISO 3166 alpha-2 country code.",
-    )
-    checksum_digits = property(
-        partial(common.Base._get_component, start=2, end=4),
-        doc="str: Two digit checksum of the IBAN.",
-    )
-    bank_code = property(
-        partial(_get_code, code_type="bank_code"),
-        doc="str: The country specific bank-code.",
-    )
-    branch_code = property(
-        partial(_get_code, code_type="branch_code"),
-        doc="str or None: The branch-code of the bank if available.",
-    )
-    account_code = property(
-        partial(_get_code, code_type="account_code"), doc="str: The customer specific account-code"
-    )
+    @property
+    def bban(self) -> str:
+        """str: The BBAN part of the IBAN."""
+        return self._get_component(start=4)
+
+    @property
+    def country_code(self) -> str:
+        """str: ISO 3166 alpha-2 country code."""
+        return self._get_component(start=0, end=2)
+
+    @property
+    def checksum_digits(self) -> str:
+        """str: Two digit checksum of the IBAN."""
+        return self._get_component(start=2, end=4)
+
+    @property
+    def bank_code(self) -> str:
+        """str: The country specific bank-code."""
+        return self._get_code(code_type="bank_code")
+
+    @property
+    def branch_code(self) -> str:
+        """str or None: The branch-code of the bank if available."""
+        return self._get_code(code_type="branch_code")
+
+    @property
+    def account_code(self) -> str:
+        """str: The customer specific account-code"""
+        return self._get_code(code_type="account_code")
 
 
-def add_bban_regex(country, spec):
+def add_bban_regex(country: str, spec: Dict) -> Dict:
     bban_spec = spec["bban_spec"]
     spec_re = r"(\d+)(!)?([{}])".format("".join(_spec_to_re.keys()))
 
-    def convert(match):
+    def convert(match: re.Match) -> str:
         quantifier = ("{%s}" if match.group(2) else "{1,%s}") % match.group(1)
         return _spec_to_re[match.group(3)] + quantifier
 
